import "dotenv/config"
import net from "net";
import { format } from "util";
import openpgp from 'openpgp'
// create this sessions PGP key
const { privateKey, publicKey } = await openpgp.generateKey({
    type: 'rsa', // Type of the key
    rsaBits: 2048, // RSA key size (defaults to 4096 bits)
    userIDs: [{ name: 'Server', email: 'neon+pgp_server__temp@hackclub.app' }], // you can pass multiple user IDs
format: 'object',
    passphrase: 'super long and hard to guess secret' // protects the private key
});
// console.log(publicKey, privateKey)
console.log(`Generate my key! ${publicKey.getFingerprint()}`)
const Commands = {
    IAC: 0xff,
    SE: 0xf0,
    NOP: 0xf1,
    DataMark: 0xf2,
    BRK: 0xf3,
    IP: 0xf4,
    AO: 0xf5,
    AYT: 0xf6,
    EC: 0xf7,
    EL: 0xf8,
    GA: 0xf9,
    SB: 0xfa,
    WILL: 0xfb,
    WONT: 0xfc,
    DO: 0xfd,
    DONT: 0xfe
};

const Options = {
    ECH: 0x1,
    SGA: 0x3,
    STS: 0x5,
    TimingMark: 0x6,
    TerminalType: 0x18,
    WindowSize: 0x1f,
    TerminalSpeed: 0x20,
    RFC: 0x21,
    LineMode: 0x22,
    EnvVars: 0x24
}

const isCommand = buffer => buffer[0] === Commands.IAC;

const isInterrput = buffer => isCommand(buffer) && buffer[1] === Commands.IP;

const sendCommand = (socket, command, option) =>
    socket.write(
        Uint8Array.from(
            [Commands.IAC, command, option].filter(b => b !== undefined)
        )
    );
//@ts-ignore
const sendEraseLine = socket => sendCommand(socket, Commands.EL);

const telnet =  {
    isInterrput,
    Commands,
    sendCommand,
    sendEraseLine,
    Options
};
const Config = {
    "server": {
        // nests ram COULD not handle more then 20 sadly 
        "maxConnections": 20,
        "noInputTimeout": 55_000 ,
        "host": "127.0.0.1",
        "port": process.env.PORT
    },
    "chat": {
        "messageLength": 150,
        "messages": {
            "welcome": "Welcome to chat server. Press CTRL+C to leave.\n(note: pgp key name will be used in server. extra metadata such as [email, pgp key id] can be found via whois cmd)",
            "nickname": "Enter your public PGP key (in base64, in one line once done write in a new line 'ENDKEY'): ",
            "logout": "%s has left the chat",
            "login": "%s has joined the chat",
            "not_allowed": "You may only connect from the nest server. This is not a public server."
        },
        "serverNickname": "PGP chat"
    }
}

const Messages = Config.chat.messages;
// to prevent my ram from being disposed 
const all_connections = [];
const clients = [];
setInterval(() => {
    all_connections.forEach(client => {
    try {
        if(Date.now() - client.lastInput >= 0) {
            client.end(`Timed out`)
        }
    } catch (e) {
        console.error(e.message, 'e')
    }
    })
}, Config.server.noInputTimeout)
const messageListener = fn => data => {
    console.log('Received data', data, data.toString());

    data = data
        .toString()
        .trim()
        .replace(/[^\w\d \t]/g, '')    
        .substring(0, Config.chat.messageLength);
    
    data && fn(data);
};

const exitHandler = socket => {
    return data => telnet.isInterrput(data) && socket.end('Bye bye.');
};

const broadcastHandler = (socket, nickname) => {
  return  messageListener(message => {
        fetch(process.env.SLACK_CHAT_WORKFLOW_URL, {
            method: 'POST',
            headers: {'Content-Type': 'application/json', 'User-Agent': 'insomnia/9.3.2'},
            body: JSON.stringify({ message, name: nickname })
          })
            .then(response => response.json())
            .catch(err => console.error(err));    
        return sendBroadcast(socket, nickname, message)
        
    });      
}

const endHandler = (socket, nickname) => () => {
    clients.splice(clients.indexOf(socket), 1);
    sendBroadcast(null, null, format(Messages.logout, nickname));

    console.log(`${socket.remoteAddress} disconnected`);
};

function sendMessageLn(socket, from, message) {
    sendMessage(socket, from, message + '\r\n');
}

function sendMessage(socket, from, message) {
    if (from) message = `<${from}> ${message}`;
    socket.write(message);
}

function sendBroadcast(socket, from, message) {
    if (!message) return;

    clients.forEach(client => {
        telnet.sendEraseLine(client);
        if (client !== socket) sendMessageLn(client, from, message);
    });
}

function askNickname(socket) {
    return new Promise((resolve, reject) => {
        let str = "";
        const ask = () => {
            socket.on('data', onData);
        if(!str.length)  sendMessage(socket, null, Messages.nickname);
        };

        const onData = messageListener(data => {
            if(data.toString().includes('ENDKEY')) {
                socket.off('data', onData)
                resolve(str);
            } else {
                str += data.toString()
                if (!data.length) return ask();
    if(!data.toString().includes('ENDKEY')) return ask()
    
            }
    
        });

        ask();
    });
}

function createServer() {
    const server = net.createServer(async socket => {

        try {
            console.log(socket.remoteAddress)
            if(![`::1`].includes(socket.remoteAddress.trim())) {
            sendMessageLn(socket, null, Messages.not_allowed);
            socket.end(`Bye`)
            return;
            }
            socket.on('data', () => {
            //@ts-ignore
all_connections[all_connections.indexOf(socket)].lastInput = Date.now()        
            })
            //@ts-ignore
            socket.lastInput = Date.now()
            all_connections.push(socket)
            console.log(`Connection from ${socket.remoteAddress}`);

            sendMessageLn(socket, null, Messages.welcome);

            socket.on('data', exitHandler(socket));

            const raw_pgp_data = await askNickname(socket);
            console.debug(`$afternic`)
            let pgp_key=null, nickname= null;
try {
    // console.log(raw_pgp_data)
    console.log(Buffer.from(raw_pgp_data as string, 'base64').toString(), Buffer.from(raw_pgp_data as string, 'base64').toString().length)
     pgp_key = await openpgp.readKey({
        armoredKey: Buffer.from(raw_pgp_data as string, 'base64').toString(),
    })
    // .then(d=>d.users[0])
    console.log(pgp_key.users[0])
    nickname = pgp_key
    console.log(`debug`)
    debugger;
} catch (e) {
    console.error(e.message)
    socket.end(`Invalid Key.`)
    return;
}
            clients.push(socket);
            socket.on('data', broadcastHandler(socket, nickname));

            // telnet.sendCommand(socket, telnet.Commands.NOP);
            // telnet.sendCommand(socket, telnet.Commands.WILL, telnet.Options.ECH);
            // telnet.sendCommand(socket, telnet.Commands.WILL, telnet.Options.SGA);

            sendBroadcast(null, null, format(Messages.login, nickname));
            console.log(
                `${socket.remoteAddress} joined the chat as ${nickname}`
            );
        } catch (e) {
            console.log('error', e);
            socket.end();
        }
    });

    server.maxConnections = Config.server.maxConnections;
    server.on('close', () => {
        console.log(`Server closed.`);
    });

    return server;
}

  const listner =   createServer().listen(parseInt(Config.server.port), () => {
    //@ts-ignore
        console.log(`Server listening on port ${listner.address().port ? listner.address().port : listner.address()}`);

    });
